==> audiochain-1.1.5-src.tgz: org/audiochain/model/AudioDeviceChain.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderAdapter;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.DelegatingAudioDataReader;
import org.audiochain.io.FrameObserver;

/**
 * This class holds the reader chain for playback and capture devices. It
 * creates and manipulates the chain at runtime. This class holds a list of
 * chainable audio devices.
 * 
 * @author Florian Paul
 */
public class AudioDeviceChain implements Serializable, BinaryDatasetProvider, ChainableAudioDeviceListener {
	private static final long serialVersionUID = 1L;
	private Collection<ChainableAudioDevice> chainableAudioDevices;
	private AudioProject audioProject;
	private boolean enabled = true;
	private String identifier = IdentifierGenerator.generateIdentifier();
	private Collection<String> removedChainableAudioDeviceIdentifiers;
	private transient Collection<AudioDeviceChainListener> listeners;
	private transient Map<AudioDevice, AudioDataReaderChainLink> audioDeviceReadersMap;
	private transient boolean reversed;
	private transient AudioDataReader sourceAudioDataReader;
	private transient float frameRate;
	private transient int sampleSizeInBits;
	private transient FrameObserver frameObserver;
	/*
	 * the create chain caller should "see" always this delegate. This enables
	 * the audio device chain to manipulate the other readers while it
	 * encapsulates the chain from the caller by the delegation reader.
	 */
	private transient AudioDataReaderChainLink baseAudioDataReaderChainLink;
	private transient final AudioDataReader nullReader = new AudioDataReaderAdapter();
	private transient AudioDataReader endReader = new AudioDataReaderAdapter() {
		@Override
		public int read(int[] data) throws IOException {
			return -1;
		}
	};
	private transient boolean syncFramePosition;

	/**
	 * For Deserialization only
	 */
	@Deprecated
	AudioDeviceChain() {

	}

	public AudioDeviceChain(AudioProject audioProject) {
		this.audioProject = audioProject;
	}

	public AudioDataReader createAudioDataReaderChain(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		return createAudioDataReaderChain(source, frameRate, sampleSizeInBits, frameObserver, false);
	}

	public AudioDataReader createReversedAudioDataReaderChain(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		return createAudioDataReaderChain(source, frameRate, sampleSizeInBits, frameObserver, true);
	}

	private synchronized AudioDataReader createAudioDataReaderChain(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver, boolean reversed) {
		this.reversed = reversed;
		this.sourceAudioDataReader = new PerformanceMonitor(source);
		this.frameRate = frameRate;
		this.sampleSizeInBits = sampleSizeInBits;
		this.frameObserver = frameObserver;
		if (audioDeviceReadersMap == null) {
			audioDeviceReadersMap = new HashMap<AudioDevice, AudioDataReaderChainLink>();
		} else {
			audioDeviceReadersMap.clear();
		}
		return updateAudioDataReaderChain();
	}

	private synchronized AudioDataReader updateAudioDataReaderChain() {
		if (audioDeviceReadersMap == null) {
			return null;
		}

		// first set for each active reader the null reader to prevent
		// inconsistency and latency
		Collection<AudioDataReaderChainLink> activeReaders = audioDeviceReadersMap.values();
		for (AudioDataReaderChainLink audioDataReaderChainLink : activeReaders) {
			audioDataReaderChainLink.setSource(nullReader);
		}

		AudioDataReader sourceReader = sourceAudioDataReader;
		if (enabled) {
			Collection<ChainableAudioDevice> audioDevices = reversed ? getReversedAudioDevices() : getAudioDevices();
			for (ChainableAudioDevice chainDevice : audioDevices) {
				chainDevice.addChainableAudioDeviceListener(this);
				if (chainDevice.isEnabled()) {
					AudioDataReaderChainLink chainLink = audioDeviceReadersMap.get(chainDevice);
					if (chainLink == null) {
						chainLink = chainDevice.createAudioDataReaderChainLink(sourceReader, frameRate, sampleSizeInBits, frameObserver);
						chainLink = new PerformanceMonitor(chainLink);
						audioDeviceReadersMap.put(chainDevice, chainLink);
					} else {
						chainLink.setSource(sourceReader);
					}
					sourceReader = chainLink;
				} else {
					// stop running readers of devices that become disabled
					closeAndRemoveAudioDataReaderChainLink(chainDevice);
				}
			}
		}
		if (baseAudioDataReaderChainLink == null) {
			baseAudioDataReaderChainLink = new DelegatingAudioDataReader(enabled ? sourceReader : endReader) {
				@Override
				public int read(int[] data) throws IOException {
					if (syncFramePosition) {
						syncFramePosition = false;
						seek(frameObserver.getReaderFramePosition());
					}
					return super.read(data);
				}

				@Override
				public void close() throws IOException {
					super.close();
					sourceAudioDataReader = null;
					if (audioDeviceReadersMap != null) {
						audioDeviceReadersMap.clear();
					}
				}
			};
		} else {
			baseAudioDataReaderChainLink.setSource(enabled ? sourceReader : endReader);
		}
		return baseAudioDataReaderChainLink;
	}

	public void addAudioDevice(ChainableAudioDevice audioDevice) {
		if (chainableAudioDevices == null) {
			chainableAudioDevices = new LinkedHashSet<ChainableAudioDevice>();
		}
		chainableAudioDevices.add(audioDevice);
		audioDevice.initialize(audioProject);
		// give any listeners a chance to register themselves to the new audio
		// device before it becomes chained into the process chain
		fireAudioDeviceAdded(audioDevice);
		updateAudioDataReaderChain();
	}

	public void removeAudioDevice(ChainableAudioDevice audioDevice) {
		if (chainableAudioDevices != null && chainableAudioDevices.contains(audioDevice)) {
			audioDevice.removeChainableAudioDeviceListener(this);
			audioDevice.destroy();
			chainableAudioDevices.remove(audioDevice);
			if (removedChainableAudioDeviceIdentifiers == null) {
				removedChainableAudioDeviceIdentifiers = new LinkedHashSet<String>();
			}
			removedChainableAudioDeviceIdentifiers.add(audioDevice.getIdentifier());
			updateAudioDataReaderChain();
			if (audioDeviceReadersMap != null) {
				closeAndRemoveAudioDataReaderChainLink(audioDevice);
			}
			fireAudioDeviceRemoved(audioDevice);
		}
	}

	public boolean isRemoved(ChainableAudioDevice audioDevice) {
		return removedChainableAudioDeviceIdentifiers != null && removedChainableAudioDeviceIdentifiers.contains(audioDevice.getIdentifier());
	}

	private void closeAndRemoveAudioDataReaderChainLink(ChainableAudioDevice audioDevice) {
		AudioDataReaderChainLink audioDataReaderChainLink = audioDeviceReadersMap.remove(audioDevice);
		if (audioDataReaderChainLink != null) {
			audioDataReaderChainLink.setSource(nullReader);
			audioDataReaderChainLink.stop();
			try {
				audioDataReaderChainLink.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public void removeAllAudioDevices() {
		if (chainableAudioDevices != null) {
			for (ChainableAudioDevice device : new LinkedHashSet<ChainableAudioDevice>(chainableAudioDevices)) {
				removeAudioDevice(device);
			}
		}
	}

	public Collection<ChainableAudioDevice> getAudioDevices() {
		Collection<ChainableAudioDevice> audioDevices = new LinkedHashSet<ChainableAudioDevice>();
		if (chainableAudioDevices != null) {
			audioDevices.addAll(chainableAudioDevices);
		}
		return Collections.unmodifiableCollection(audioDevices);
	}

	public Collection<ChainableAudioDevice> getReversedAudioDevices() {
		List<ChainableAudioDevice> reversed = new ArrayList<ChainableAudioDevice>(getAudioDevices());
		Collections.reverse(reversed);
		return Collections.unmodifiableCollection(reversed);
	}

	public void decrementAudioDevicePositionIndex(ChainableAudioDevice audioDevice) {
		moveAudioDevicePosition(audioDevice, false);
	}

	public void incrementAudioDevicePositionIndex(ChainableAudioDevice audioDevice) {
		moveAudioDevicePosition(audioDevice, true);
	}

	private void moveAudioDevicePosition(ChainableAudioDevice audioDevice, boolean increment) {
		if (chainableAudioDevices != null && chainableAudioDevices.contains(audioDevice)) {
			if (movePositionIndex(audioDevice, chainableAudioDevices, increment)) {
				updateAudioDataReaderChain();
				fireAudioDeviceMoved(audioDevice, increment);
			}
		}
	}

	/**
	 * @return true if the object was moved on its index or false if nothing was
	 *         changed
	 */
	static <E> boolean movePositionIndex(E object, Collection<E> elements, boolean increment) {
		if (elements == null) {
			return false;
		}
		if (!elements.contains(object)) {
			throw new IllegalArgumentException("The collection does not contain the given object '" + object + "'.");
		}
		List<E> list = new ArrayList<E>(elements);
		int oldIndex = list.indexOf(object);
		if (increment) {
			if (oldIndex < list.size() - 1) {
				list.remove(object);
				int newIndex = oldIndex + 1;
				list.add(newIndex, object);
				elements.clear();
				elements.addAll(list);
				return true;
			}
		} else {
			if (oldIndex > 0) {
				list.remove(object);
				int newIndex = oldIndex - 1;
				list.add(newIndex, object);
				elements.clear();
				elements.addAll(list);
				return true;
			}
		}
		return false;
	}

	private void fireAudioDeviceAdded(AudioDevice audioDevice) {
		if (listeners == null) {
			return;
		}
		for (AudioDeviceChainListener listener : listeners) {
			listener.audioDeviceAdded(audioDevice, this);
		}
	}

	private void fireAudioDeviceRemoved(AudioDevice audioDevice) {
		if (listeners == null) {
			return;
		}
		for (AudioDeviceChainListener listener : listeners) {
			listener.audioDeviceRemoved(audioDevice, this);
		}
	}

	private void fireAudioDeviceMoved(AudioDevice audioDevice, boolean increment) {
		if (listeners == null) {
			return;
		}
		for (AudioDeviceChainListener listener : listeners) {
			listener.audioDeviceMoved(audioDevice, this, increment);
		}
	}

	public void addAudioDeviceChainListener(AudioDeviceChainListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<AudioDeviceChainListener>();
		}
		listeners.add(listener);
	}

	public void removeAudioDeviceChainListener(AudioDeviceChainListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	@Override
	public void audioDeviceEnabledChanged(ChainableAudioDevice chainableAudioDevice, boolean oldValue, boolean newValue) {
		updateAudioDataReaderChain();
	}

	void setEnabled(boolean enabled) {
		syncFramePosition = !this.enabled && enabled;
		this.enabled = enabled;
		updateAudioDataReaderChain();
	}

	@Override
	public String getIdentifier() {
		return identifier;
	}

	public void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	ChainableAudioDevice findChainableAudioDeviceByIdentifier(String identifier) {
		Collection<ChainableAudioDevice> audioDevices = getAudioDevices();
		for (ChainableAudioDevice chainableAudioDevice : audioDevices) {
			if (chainableAudioDevice.getIdentifier().equals(identifier)) {
				return chainableAudioDevice;
			}
		}
		return null;
	}

	AudioProject getAudioProject() {
		return audioProject;
	}

	void setAudioProject(AudioProject audioProject) {
		this.audioProject = audioProject;
	}

	@Override
	public void synchronizeBy(Synchronizable source) throws InstantiationException, IllegalAccessException {
		if (source instanceof AudioDeviceChain && source.getIdentifier().equals(getIdentifier())) {
			AudioDeviceChain sourceAudioDeviceChain = (AudioDeviceChain) source;
			Collection<ChainableAudioDevice> sourceChainableAudioDevices = sourceAudioDeviceChain.getAudioDevices();
			for (ChainableAudioDevice sourceChainableAudioDevice : sourceChainableAudioDevices) {
				if (isRemoved(sourceChainableAudioDevice)) {
					// ignore this device
					continue;
				}
				String sourceChainableAudioDeviceIdentifier = sourceChainableAudioDevice.getIdentifier();
				ChainableAudioDevice targetChainableAudioDevice = findChainableAudioDeviceByIdentifier(sourceChainableAudioDeviceIdentifier);
				if (targetChainableAudioDevice == null) {
					if (sourceChainableAudioDevice instanceof Copyable) {
						targetChainableAudioDevice = (ChainableAudioDevice) ((Copyable<?>) sourceChainableAudioDevice).copy();
					} else {
						targetChainableAudioDevice = sourceChainableAudioDevice.getClass().newInstance();
					}
					targetChainableAudioDevice.setIdentifier(sourceChainableAudioDeviceIdentifier);
					// this status is only synced on new instances because it is
					// user specific and has no further relevance
					targetChainableAudioDevice.getUserInterfaceContext().setHidden(sourceChainableAudioDevice.getUserInterfaceContext().isHidden());
					addAudioDevice(targetChainableAudioDevice);
				}
				targetChainableAudioDevice.synchronizeBy(sourceChainableAudioDevice);
			}
			Collection<ChainableAudioDevice> targetChainableAudioDevices = getAudioDevices();
			for (ChainableAudioDevice targetChainableAudioDevice : new LinkedHashSet<ChainableAudioDevice>(targetChainableAudioDevices)) {
				if (sourceAudioDeviceChain.isRemoved(targetChainableAudioDevice)) {
					removeAudioDevice(targetChainableAudioDevice);
				}
			}
		}
	}

	@Override
	public Collection<BinaryDataset> getAllBinaryDatasets() {
		Collection<BinaryDataset> ret = new HashSet<BinaryDataset>();
		Collection<ChainableAudioDevice> audioDevices = getAudioDevices();
		for (ChainableAudioDevice chainableAudioDevice : audioDevices) {
			if (chainableAudioDevice instanceof BinaryDatasetProvider) {
				BinaryDatasetProvider dev = (BinaryDatasetProvider) chainableAudioDevice;
				ret.addAll(dev.getAllBinaryDatasets());
			}
		}
		return ret;
	}

	Monitorable getSourceMonitor() {
		return (Monitorable) sourceAudioDataReader;
	}

	Monitorable findMonitor(AudioDevice device) {
		if (audioDeviceReadersMap == null) {
			return null;
		}
		AudioDataReaderChainLink audioDataReaderChainLink = audioDeviceReadersMap.get(device);
		if (audioDataReaderChainLink instanceof Monitorable) {
			return (Monitorable) audioDataReaderChainLink;
		}
		return null;
	}
}

==> audiochain-1.1.6-src.tgz: org/audiochain/model/AudioDeviceChain.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.model;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderAdapter;
import org.audiochain.io.AudioDataReaderChainLink;
import org.audiochain.io.DelegatingAudioDataReader;
import org.audiochain.io.FrameObserver;

/**
 * This class holds the reader chain for playback and capture devices. It
 * creates and manipulates the chain at runtime. This class holds a list of
 * chainable audio devices.
 * 
 * @author Florian Paul
 */
public class AudioDeviceChain implements Serializable, BinaryDatasetProvider, ChainableAudioDeviceListener {
	private static final long serialVersionUID = 1L;
	private Collection<ChainableAudioDevice> chainableAudioDevices;
	private AudioProject audioProject;
	private boolean enabled = true;
	private String identifier = IdentifierGenerator.generateIdentifier();
	private Collection<String> removedChainableAudioDeviceIdentifiers;
	private transient Collection<AudioDeviceChainListener> listeners;
	private transient Map<AudioDevice, AudioDataReaderChainLink> audioDeviceReadersMap;
	private transient boolean reversed;
	private transient AudioDataReader sourceAudioDataReader;
	private transient float frameRate;
	private transient int sampleSizeInBits;
	private transient FrameObserver frameObserver;
	/*
	 * the create chain caller should "see" always this delegate. This enables
	 * the audio device chain to manipulate the other readers while it
	 * encapsulates the chain from the caller by the delegation reader.
	 */
	private transient AudioDataReaderChainLink baseAudioDataReaderChainLink;
	private transient final AudioDataReader nullReader = new AudioDataReaderAdapter();
	private transient AudioDataReader endReader = new AudioDataReaderAdapter() {
		@Override
		public int read(int[] data) throws IOException {
			return -1;
		}
	};
	private transient boolean syncFramePosition;

	/**
	 * For Deserialization only
	 */
	@Deprecated
	AudioDeviceChain() {

	}

	public AudioDeviceChain(AudioProject audioProject) {
		this.audioProject = audioProject;
	}

	public AudioDataReader createAudioDataReaderChain(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		return createAudioDataReaderChain(source, frameRate, sampleSizeInBits, frameObserver, false);
	}

	public AudioDataReader createReversedAudioDataReaderChain(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver) {
		return createAudioDataReaderChain(source, frameRate, sampleSizeInBits, frameObserver, true);
	}

	private synchronized AudioDataReader createAudioDataReaderChain(AudioDataReader source, float frameRate, int sampleSizeInBits, FrameObserver frameObserver, boolean reversed) {
		this.reversed = reversed;
		this.sourceAudioDataReader = new PerformanceMonitor(source);
		this.frameRate = frameRate;
		this.sampleSizeInBits = sampleSizeInBits;
		this.frameObserver = frameObserver;
		if (audioDeviceReadersMap == null) {
			audioDeviceReadersMap = new HashMap<AudioDevice, AudioDataReaderChainLink>();
		} else {
			audioDeviceReadersMap.clear();
		}
		return updateAudioDataReaderChain();
	}

	private synchronized AudioDataReader updateAudioDataReaderChain() {
		if (audioDeviceReadersMap == null) {
			return null;
		}

		// first set for each active reader the null reader to prevent
		// inconsistency and latency
		Collection<AudioDataReaderChainLink> activeReaders = audioDeviceReadersMap.values();
		for (AudioDataReaderChainLink audioDataReaderChainLink : activeReaders) {
			audioDataReaderChainLink.setSource(nullReader);
		}

		AudioDataReader sourceReader = sourceAudioDataReader;
		if (enabled) {
			Collection<ChainableAudioDevice> audioDevices = reversed ? getReversedAudioDevices() : getAudioDevices();
			for (ChainableAudioDevice chainDevice : audioDevices) {
				chainDevice.addChainableAudioDeviceListener(this);
				if (chainDevice.isEnabled()) {
					AudioDataReaderChainLink chainLink = audioDeviceReadersMap.get(chainDevice);
					if (chainLink == null) {
						chainLink = chainDevice.createAudioDataReaderChainLink(sourceReader, frameRate, sampleSizeInBits, frameObserver);
						chainLink = new PerformanceMonitor(chainLink);
						audioDeviceReadersMap.put(chainDevice, chainLink);
					} else {
						chainLink.setSource(sourceReader);
					}
					sourceReader = chainLink;
				} else {
					// stop running readers of devices that become disabled
					closeAndRemoveAudioDataReaderChainLink(chainDevice);
				}
			}
		}
		if (baseAudioDataReaderChainLink == null) {
			baseAudioDataReaderChainLink = new DelegatingAudioDataReader(enabled ? sourceReader : endReader) {
				@Override
				public int read(int[] data) throws IOException {
					if (syncFramePosition) {
						syncFramePosition = false;
						seek(frameObserver.getReaderFramePosition());
					}
					return super.read(data);
				}

				@Override
				public void close() throws IOException {
					super.close();
					sourceAudioDataReader = null;
					if (audioDeviceReadersMap != null) {
						audioDeviceReadersMap.clear();
					}
				}
			};
		} else {
			baseAudioDataReaderChainLink.setSource(enabled ? sourceReader : endReader);
		}
		return baseAudioDataReaderChainLink;
	}

	public void addAudioDevice(ChainableAudioDevice audioDevice) {
		if (chainableAudioDevices == null) {
			chainableAudioDevices = new LinkedHashSet<ChainableAudioDevice>();
		}
		chainableAudioDevices.add(audioDevice);
		audioDevice.initialize(audioProject);
		// give any listeners a chance to register themselves to the new audio
		// device before it becomes chained into the process chain
		fireAudioDeviceAdded(audioDevice);
		updateAudioDataReaderChain();
	}

	public void removeAudioDevice(ChainableAudioDevice audioDevice) {
		if (chainableAudioDevices != null && chainableAudioDevices.contains(audioDevice)) {
			audioDevice.removeChainableAudioDeviceListener(this);
			audioDevice.destroy();
			chainableAudioDevices.remove(audioDevice);
			if (removedChainableAudioDeviceIdentifiers == null) {
				removedChainableAudioDeviceIdentifiers = new LinkedHashSet<String>();
			}
			removedChainableAudioDeviceIdentifiers.add(audioDevice.getIdentifier());
			updateAudioDataReaderChain();
			if (audioDeviceReadersMap != null) {
				closeAndRemoveAudioDataReaderChainLink(audioDevice);
			}
			fireAudioDeviceRemoved(audioDevice);
		}
	}

	public boolean isRemoved(ChainableAudioDevice audioDevice) {
		return removedChainableAudioDeviceIdentifiers != null && removedChainableAudioDeviceIdentifiers.contains(audioDevice.getIdentifier());
	}

	private void closeAndRemoveAudioDataReaderChainLink(ChainableAudioDevice audioDevice) {
		AudioDataReaderChainLink audioDataReaderChainLink = audioDeviceReadersMap.remove(audioDevice);
		if (audioDataReaderChainLink != null) {
			audioDataReaderChainLink.setSource(nullReader);
			audioDataReaderChainLink.stop();
			try {
				audioDataReaderChainLink.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	public void removeAllAudioDevices() {
		if (chainableAudioDevices != null) {
			for (ChainableAudioDevice device : new LinkedHashSet<ChainableAudioDevice>(chainableAudioDevices)) {
				removeAudioDevice(device);
			}
		}
	}

	public Collection<ChainableAudioDevice> getAudioDevices() {
		Collection<ChainableAudioDevice> audioDevices = new LinkedHashSet<ChainableAudioDevice>();
		if (chainableAudioDevices != null) {
			audioDevices.addAll(chainableAudioDevices);
		}
		return Collections.unmodifiableCollection(audioDevices);
	}

	public Collection<ChainableAudioDevice> getReversedAudioDevices() {
		List<ChainableAudioDevice> reversed = new ArrayList<ChainableAudioDevice>(getAudioDevices());
		Collections.reverse(reversed);
		return Collections.unmodifiableCollection(reversed);
	}

	public void decrementAudioDevicePositionIndex(ChainableAudioDevice audioDevice) {
		moveAudioDevicePosition(audioDevice, false);
	}

	public void incrementAudioDevicePositionIndex(ChainableAudioDevice audioDevice) {
		moveAudioDevicePosition(audioDevice, true);
	}

	private void moveAudioDevicePosition(ChainableAudioDevice audioDevice, boolean increment) {
		if (chainableAudioDevices != null && chainableAudioDevices.contains(audioDevice)) {
			if (movePositionIndex(audioDevice, chainableAudioDevices, increment)) {
				updateAudioDataReaderChain();
				fireAudioDeviceMoved(audioDevice, increment);
			}
		}
	}

	/**
	 * @return true if the object was moved on its index or false if nothing was
	 *         changed
	 */
	public static <E> boolean movePositionIndex(E object, Collection<E> elements, boolean increment) {
		if (elements == null) {
			return false;
		}
		if (!elements.contains(object)) {
			throw new IllegalArgumentException("The collection does not contain the given object '" + object + "'.");
		}
		List<E> list = new ArrayList<E>(elements);
		int oldIndex = list.indexOf(object);
		if (increment) {
			if (oldIndex < list.size() - 1) {
				list.remove(object);
				int newIndex = oldIndex + 1;
				list.add(newIndex, object);
				elements.clear();
				elements.addAll(list);
				return true;
			}
		} else {
			if (oldIndex > 0) {
				list.remove(object);
				int newIndex = oldIndex - 1;
				list.add(newIndex, object);
				elements.clear();
				elements.addAll(list);
				return true;
			}
		}
		return false;
	}

	private void fireAudioDeviceAdded(AudioDevice audioDevice) {
		if (listeners == null) {
			return;
		}
		for (AudioDeviceChainListener listener : listeners) {
			listener.audioDeviceAdded(audioDevice, this);
		}
	}

	private void fireAudioDeviceRemoved(AudioDevice audioDevice) {
		if (listeners == null) {
			return;
		}
		for (AudioDeviceChainListener listener : listeners) {
			listener.audioDeviceRemoved(audioDevice, this);
		}
	}

	private void fireAudioDeviceMoved(AudioDevice audioDevice, boolean increment) {
		if (listeners == null) {
			return;
		}
		for (AudioDeviceChainListener listener : listeners) {
			listener.audioDeviceMoved(audioDevice, this, increment);
		}
	}

	public void addAudioDeviceChainListener(AudioDeviceChainListener listener) {
		if (listeners == null) {
			listeners = new LinkedHashSet<AudioDeviceChainListener>();
		}
		listeners.add(listener);
	}

	public void removeAudioDeviceChainListener(AudioDeviceChainListener listener) {
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
	}

	@Override
	public void audioDeviceEnabledChanged(ChainableAudioDevice chainableAudioDevice, boolean oldValue, boolean newValue) {
		updateAudioDataReaderChain();
	}

	void setEnabled(boolean enabled) {
		syncFramePosition = !this.enabled && enabled;
		this.enabled = enabled;
		updateAudioDataReaderChain();
	}

	@Override
	public String getIdentifier() {
		return identifier;
	}

	public void setIdentifier(String identifier) {
		this.identifier = identifier;
	}

	ChainableAudioDevice findChainableAudioDeviceByIdentifier(String identifier) {
		Collection<ChainableAudioDevice> audioDevices = getAudioDevices();
		for (ChainableAudioDevice chainableAudioDevice : audioDevices) {
			if (chainableAudioDevice.getIdentifier().equals(identifier)) {
				return chainableAudioDevice;
			}
		}
		return null;
	}

	AudioProject getAudioProject() {
		return audioProject;
	}

	void setAudioProject(AudioProject audioProject) {
		this.audioProject = audioProject;
	}

	@Override
	public void synchronizeBy(Synchronizable source) throws InstantiationException, IllegalAccessException {
		if (source instanceof AudioDeviceChain && source.getIdentifier().equals(getIdentifier())) {
			AudioDeviceChain sourceAudioDeviceChain = (AudioDeviceChain) source;
			Collection<ChainableAudioDevice> sourceChainableAudioDevices = sourceAudioDeviceChain.getAudioDevices();
			for (ChainableAudioDevice sourceChainableAudioDevice : sourceChainableAudioDevices) {
				if (isRemoved(sourceChainableAudioDevice)) {
					// ignore this device
					continue;
				}
				String sourceChainableAudioDeviceIdentifier = sourceChainableAudioDevice.getIdentifier();
				ChainableAudioDevice targetChainableAudioDevice = findChainableAudioDeviceByIdentifier(sourceChainableAudioDeviceIdentifier);
				if (targetChainableAudioDevice == null) {
					if (sourceChainableAudioDevice instanceof Copyable) {
						targetChainableAudioDevice = (ChainableAudioDevice) ((Copyable<?>) sourceChainableAudioDevice).copy();
					} else {
						targetChainableAudioDevice = sourceChainableAudioDevice.getClass().newInstance();
					}
					targetChainableAudioDevice.setIdentifier(sourceChainableAudioDeviceIdentifier);
					// this status is only synced on new instances because it is
					// user specific and has no further relevance
					targetChainableAudioDevice.getUserInterfaceContext().setHidden(sourceChainableAudioDevice.getUserInterfaceContext().isHidden());
					addAudioDevice(targetChainableAudioDevice);
				}
				targetChainableAudioDevice.synchronizeBy(sourceChainableAudioDevice);
			}
			Collection<ChainableAudioDevice> targetChainableAudioDevices = getAudioDevices();
			for (ChainableAudioDevice targetChainableAudioDevice : new LinkedHashSet<ChainableAudioDevice>(targetChainableAudioDevices)) {
				if (sourceAudioDeviceChain.isRemoved(targetChainableAudioDevice)) {
					removeAudioDevice(targetChainableAudioDevice);
				}
			}
		}
	}

	@Override
	public Collection<BinaryDataset> getAllBinaryDatasets() {
		Collection<BinaryDataset> ret = new HashSet<BinaryDataset>();
		Collection<ChainableAudioDevice> audioDevices = getAudioDevices();
		for (ChainableAudioDevice chainableAudioDevice : audioDevices) {
			if (chainableAudioDevice instanceof BinaryDatasetProvider) {
				BinaryDatasetProvider dev = (BinaryDatasetProvider) chainableAudioDevice;
				ret.addAll(dev.getAllBinaryDatasets());
			}
		}
		return ret;
	}

	Monitorable getSourceMonitor() {
		return (Monitorable) sourceAudioDataReader;
	}

	Monitorable findMonitor(AudioDevice device) {
		if (audioDeviceReadersMap == null) {
			return null;
		}
		AudioDataReaderChainLink audioDataReaderChainLink = audioDeviceReadersMap.get(device);
		if (audioDataReaderChainLink instanceof Monitorable) {
			return (Monitorable) audioDataReaderChainLink;
		}
		return null;
	}
}
